module Tree where

import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Servant.Client hiding (Response)
import Control.Lens.Getter ((^.))
import Data.Map ((!?), (!), insert, empty, Map, fromList, toList)
import qualified Data.Map as M
import Data.Map.Internal (merge, preserveMissing, zipWithMatched)
import Data.Aeson
import Data.List (groupBy, isInfixOf)
import Data.List.Split (chunksOf)
import Data.Either (rights)
import Data.Maybe (fromMaybe)
import Data.Scientific (Scientific, floatingOrInteger)
import qualified Data.Vector as V
import Text.Printf
import GHC.Generics

import HAL.Client
import HAL
import HAL.Doc.Struct
import Data.Text (Text, pack)

formatParentIdRequest :: [Struct] -> Maybe Text
formatParentIdRequest [] = Nothing
formatParentIdRequest (x:[]) = Just . pack . show $ _struct_docid x
formatParentIdRequest (x:xs) =
  (Just . pack . show $ _struct_docid x)
  <> (Just " || ")
  <> formatParentIdRequest xs


ds2Child :: [Struct] -> IO [Struct]
ds2Child ds =  do
  rs <- sequence $ runStructureRequest <$> formatedRequests ds
  return . concat $ (^. docs) <$> rights rs
  where formatedRequest docs' = Just "parentDocid_i:(" <> formatParentIdRequest docs' <> Just ")"
        formatedRequests docs'' = formatedRequest <$> chunksOf 100 docs''

fetchChildren :: [Struct] -> IO [Struct]
fetchChildren [] = pure []
fetchChildren ds = (ds <>) <$> (fetchChildren =<< ds2Child ds)

isChildOf :: Struct -> Struct -> Bool
isChildOf (Struct i l p) (Struct i' l' p') = not . null $ filter (\id -> id == (pack $ show i)) p'

data DocTree = DocTree Struct Int [DocTree]
  deriving (Show, Generic)

buildTree :: Int -> [Struct] -> Struct -> DocTree
buildTree depth docs id = DocTree id depth (buildTree (depth + 1) docs <$> children)
  where children = filter (isChildOf id) docs

formatTree :: DocTree -> String
formatTree tree@(DocTree doc depth children) =
  printf "%s%s\n%s" (addSpace) (show doc) (concat $ formatTree <$> children)
  where addSpace
          | depth > 0 = "├" <> (replicate (depth * depth + depth) '─')
          | otherwise = "🌲"

findDeepest :: Map Int Int -> DocTree -> Map Int Int
findDeepest m tree@(DocTree doc@(Struct docid _ _) depth children) =
  mergeMap map1 maps
  where map1 = insert docid depth m
        maps = foldl mergeMap empty $ findDeepest m <$> children
        mergeMap m1 m2 = merge preserveMissing preserveMissing (zipWithMatched whenMatch) m1 m2
        whenMatch k v v' = if v > v' then v else v'

isDeep :: Map Int Int -> DocTree -> Bool
isDeep m (DocTree (Struct id _ _) depth _) = depth >= (fromMaybe 0 $ M.lookup id m)

removeDuplicate :: Map Int Int -> DocTree -> DocTree
removeDuplicate deepMap tree@(DocTree doc depth children) = DocTree doc depth (removeDuplicate deepMap <$> filter (isDeep deepMap) children)

noDuplicateTree :: DocTree -> DocTree
noDuplicateTree tree = removeDuplicate (findDeepest M.empty tree) tree
